###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import copy

from vistrails.core import debug
from vistrails.core.log.workflow_exec import WorkflowExec
from vistrails.core.log.module_exec import ModuleExec
from vistrails.core.log.loop_exec import LoopExec, LoopIteration
from vistrails.core.log.group_exec import GroupExec
from vistrails.core.log.machine import Machine
from vistrails.core.modules.sub_module import Group, Abstraction
from vistrails.core.vistrail.annotation import Annotation
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.vistrail.vistrail import Vistrail
import vistrails.core.system


@apply
class DummyLogController(object):
    """DummyLogger is a class that has the entire interface for a logger
    but simply ignores the calls."""
    def start_workflow_execution(self, *args, **kwargs): return self
    def recursing(self, *args, **kwargs): return self
    def finish_workflow_execution(self, *args, **kwargs): pass
    def add_exec(self, *args, **kwargs): pass
    def start_execution(self, *args, **kwargs): pass
    def start_loop_execution(self, *args, **kwargs): return self
    def finish_loop_execution(self, *args, **kwargs): pass
    def start_iteration(self, *args, **kwargs): pass
    def finish_iteration(self, *args, **kwargs): pass
    def finish_execution(self, *args, **kwargs): pass
    def insert_module_annotations(self, *args, **kwargs): pass
    def insert_workflow_exec_annotations(self, *args, **kwargs): pass
    def add_machine(self, *args, **kwargs): return -1
    def get_iteration_from_module(self, *args, **kwargs): return None
    def __call__(self): return self


class LogController(object):
    """The top-level log controller.

    This holds a log.
    """
    _local_machine = None

    @classmethod
    def get_local_machine(cls):
        if cls._local_machine is None:
            cls._local_machine = Machine(
                    id=-1,
                    name=vistrails.core.system.current_machine(),
                    os=vistrails.core.system.systemType,
                    architecture=vistrails.core.system.current_architecture(),
                    processor=vistrails.core.system.current_processor(),
                    ram=vistrails.core.system.guess_total_memory())
        return copy.copy(cls._local_machine)

    def __init__(self, log, machine=None):
        self.log = log
        self.module_execs = {}      # vistrails_module -> *Exec
        self.parent_execs = {}      # vistrails_module -> *Exec
        self.children_execs = {}    # vistrails_module -> [*Exec]
        if machine is not None:
            self.machine = machine
        else:
            self.machine = copy.copy(self.get_local_machine())
            self.machine.id = self.log.id_scope.getNewId(Machine.vtType)

    def _create_module_exec(self, module, module_id, module_name,
                            cached):
        m_exec_id = self.log.id_scope.getNewId(ModuleExec.vtType)
        module_exec = ModuleExec(id=m_exec_id,
                                 machine_id=self.machine.id,
                                 module_id=module_id,
                                 module_name=module_name,
                                 cached=cached,
                                 ts_start=vistrails.core.system.current_time(),
                                 completed=0)
        return module_exec

    def _create_group_exec(self, group, module_id, group_name, cached):
        g_exec_id = self.log.id_scope.getNewId(GroupExec.vtType)
        if isinstance(group, Abstraction):
            group_type = 'SubWorkflow'
        else:
            group_type = 'Group'
        group_exec = GroupExec(id=g_exec_id,
                               machine_id=self.machine.id,
                               module_id=module_id,
                               group_name=group_name,
                               group_type=group_type,
                               cached=cached,
                               ts_start=vistrails.core.system.current_time(),
                               completed=0)
        return group_exec

    def _create_loop_exec(self):
        l_exec_id = self.log.id_scope.getNewId(LoopExec.vtType)
        loop_exec = LoopExec(id=l_exec_id,
                             ts_start=vistrails.core.system.current_time())
        return loop_exec

    def start_workflow_execution(self, parent_exec,
                                 vistrail=None, pipeline=None,
                                 currentVersion=None):
        """Signals the start of the execution of a pipeline.
        """
        return LogWorkflowExecController(self.log, self.machine, parent_exec,
                                         vistrail, pipeline, currentVersion)


class LogLoopController(object):
    def __init__(self, controller, loop_exec, loop_module):
        self.controller = controller
        self.loop_exec = loop_exec
        self.loop_module = loop_module

    def _create_loop_iteration(self, iteration):
        l_iteration_id = self.controller.log.id_scope.getNewId(
                LoopIteration.vtType)
        loop_iteration = LoopIteration(id=l_iteration_id,
                                       ts_start=vistrails.core.system.current_time(),
                                       iteration=iteration)
        return loop_iteration

    def finish_loop_execution(self):
        """Signals that we are done looping.
        """
        self.loop_exec.ts_end = vistrails.core.system.current_time()
        try:
            execs = self.controller.children_execs[self.loop_module]
            execs.discard(self.loop_exec)
        except KeyError:
            pass

    def start_iteration(self, looped_module, iteration):
        """Signals that we are executing a module as an iteration of the loop.
        """
        loop_iteration = self._create_loop_iteration(iteration)
        self.loop_exec.add_loop_iteration(loop_iteration)
        self.controller.parent_execs[looped_module] = loop_iteration

    def finish_iteration(self, looped_module):
        """Signals that the iteration is done.
        """
        loop_iteration = self.controller.parent_execs.get(looped_module)
        assert loop_iteration is not None

        loop_iteration.ts_end = vistrails.core.system.current_time()
        loop_iteration.completed = 1


class LogWorkflowController(LogController):
    """A log controller for a specific workflow execution.

    You get one of these by calling LogController#start_workflow_execution() or
    LogWorkflowController#recursing(). You can then add execution items through
    it.

    How does this work:
      * the interpreter sets a 'logging' attribute on summoned Module objects
        before starting the execution. Through it, the module can log events.
      * the interpreter directly uses the logger to record exceptions from the
        pipeline
      * while they are executing, the logger keeps a table mapping a Module to:
         - module_execs has the execution entry for that module
         - parent_execs is set via other modules to remember where the soon-to-
           be-created module_exec should be added (for example, loop modules
           and Group sets parent_exec on their dependent modules). If
           parent_exec is not set, the parent exec or WorkflowExec will be used
         - children_exec is a list of ongoing exec items, children of
           module_exec, such as LoopExec. They are kept their to be marked as
           finished with the same error as the module if it fails before they
           end
    """
    def __init__(self, log, machine, parent_exec, workflow_exec):
        super(LogWorkflowController, self).__init__(log, machine)
        self.parent_exec = parent_exec
        self.workflow_exec = workflow_exec

    def recursing(self, parent_exec):
        """Enters a recursing execution.

        This returns a new log controller object for that execution context.
        """
        if parent_exec in self.module_execs:
            parent_exec = self.module_execs[parent_exec]
        return LogWorkflowController(self.log, self.machine, parent_exec,
                                     self.workflow_exec)

    def get_iteration_from_module(self, module):
        """If executing this module as part of a loop, gets the iteration;

        Else returns None. Used by the interpreter to know what failed when
        getting an exception from a module.
        """
        try:
            return self.parent_execs[module].iteration
        except KeyError:
            return None

    def start_execution(self, module, module_id, module_name, cached=0):
        """Signals the start of the execution of a module (before compute).
        """
        if isinstance(module, Group):
            module_exec = self._create_group_exec(module, module_id,
                                                 module_name, cached)
        else:
            module_exec = self._create_module_exec(module, module_id,
                                                   module_name, cached)
        if module in self.module_execs is not None:
            debug.warning(
                    "%s#start_execution(module=%r, module_id=%r, "
                    "module_name=%r, cached=%r): module already has a "
                    "module_exec! Overwriting" % (
                    type(self).__name__,
                    module, module_id, module_name, cached))
        self.module_execs[module] = module_exec
        for parent_exec in (self.parent_execs.get(module), self.parent_exec,
                            self.workflow_exec):
            if parent_exec is not None:
                parent_exec.add_item_exec(module_exec)
                return
        assert False

    def start_loop_execution(self, loop_module, total_iterations=None):
        """Starts a loop.
        """
        loop_exec = self._create_loop_exec()
        for parent_exec in (self.module_execs.get(loop_module),
                            self.parent_exec):
            if parent_exec is not None:
                if isinstance(parent_exec, GroupExec):
                    parent_exec.add_item_exec(loop_exec)
                else:
                    parent_exec.add_loop_exec(loop_exec)
                break
        else:
            self.workflow_exec.add_item_exec(loop_exec)
        self.children_execs.setdefault(loop_module, set()).add(loop_exec)
        return LogLoopController(self, loop_exec, loop_module)

    def finish_execution(self, module, error, errorTrace=None, suspended=False):
        """Signals the end of the execution of a module.

        Called by a module after succeeded of suspended, or called by the
        interpreter after an exception.
        """
        module_exec = self.module_execs.pop(module, None)
        if module_exec is None:
            # The module can finish execution without starting (if it was
            # suspended, etc...)
            return
        module_exec.ts_end = vistrails.core.system.current_time()
        if suspended:
            module_exec.completed = -2
            module_exec.error = error
        elif error:
            module_exec.completed = -1
            module_exec.error = error
            if errorTrace:
                a_id = self.log.id_scope.getNewId(Annotation.vtType)
                annotation = Annotation(id=a_id,
                                        key="errorTrace",
                                        value=errorTrace)
                module_exec.add_annotation(annotation)
        else:
            module_exec.completed = 1

        for child in self.children_execs.pop(module, ()):
            child.ts_end = vistrails.core.system.current_time()
            if suspended:
                child.completed = -2
                child.error = error
            elif not error:
                child.completed = 1
            else:
                child.completed = -1
                child.error = error

    def insert_module_annotations(self, module, a_dict):
        """Adds an annotation on the execution object for this module.
        """
        for k, v in a_dict.iteritems():
            a_id = self.log.id_scope.getNewId(Annotation.vtType)
            annotation = Annotation(id=a_id,
                                    key=k,
                                    value=v)
            self.module_execs[module].add_annotation(annotation)

    def insert_workflow_exec_annotations(self, a_dict):
        """Adds an annotation on the whole workflow log object.

        The information is not associated with the execution of a specific
        module but of the whole pipeline.
        """
        if self.workflow_exec:
            for k, v in a_dict.iteritems():
                a_id = self.log.id_scope.getNewId(Annotation.vtType)
                annotation = Annotation(id=a_id,
                                        key=k,
                                        value=v)
                self.workflow_exec.add_annotation(annotation)

    def add_machine(self, machine):
        machine.id = self.log.id_scope.getNewId(Machine.vtType)
        self.workflow_exec.add_machine(machine)
        return machine.id

    def add_exec(self, exec_):
        self.workflow_exec.add_item_exec(exec_)


class LogWorkflowExecController(LogWorkflowController):
    """Top-level LogWorkflowController, returned by start_workflow_execution().

    This one has finish_workflow_execution(). The LogWorkflowController,
    obtained through recursing(), don't.
    """
    def __init__(self, log, machine, parent_exec, vistrail=None, pipeline=None,
                 currentVersion=None):
        if vistrail is not None:
            parent_type = Vistrail.vtType
            parent_id = vistrail.id
        else:
            parent_type = Pipeline.vtType
            parent_id = pipeline.id

        wf_exec_id = log.id_scope.getNewId(WorkflowExec.vtType)
        if vistrail is not None:
            session = vistrail.current_session
        else:
            session = None
        workflow_exec = WorkflowExec(
                id=wf_exec_id,
                user=vistrails.core.system.current_user(),
                ip=vistrails.core.system.current_ip(),
                vt_version=vistrails.core.system.vistrails_version(),
                ts_start=vistrails.core.system.current_time(),
                parent_type=parent_type,
                parent_id=parent_id,
                parent_version=currentVersion,
                completed=0,
                session=session,
                machines=[machine])
        log.add_workflow_exec(workflow_exec)

        super(LogWorkflowExecController, self).__init__(log, machine, parent_exec, workflow_exec)

    def finish_workflow_execution(self, errors, suspended=False):
        """Signals the end of the execution of a pipeline.
        """
        self.workflow_exec.ts_end = vistrails.core.system.current_time()
        if suspended:
            self.workflow_exec.completed = -2
        elif len(errors) > 0:
            self.workflow_exec.completed = -1
        else:
            self.workflow_exec.completed = 1



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED B Y THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import errno
import json
import os
import shutil
import subprocess
import sys

from vistrails.core.modules.vistrails_module import Module, ModuleError, IncompleteImplementation, new_module
import vistrails.core.modules.module_registry
from vistrails.core import debug
from vistrails.core.packagemanager import get_package_manager
import vistrails.core.system
from vistrails.core.system import packages_directory, vistrails_root_directory

import identifiers


cl_tools = {}


class CLTools(Module):
    """ CLTools is the base Module.
     We will create a SUDSWebService Module for each method published by 
     the web service.

    """
    def compute(self):
        raise IncompleteImplementation # pragma: no cover


SUFFIX = '.clt'
DEFAULTFILESUFFIX = '.cld'


def _eintr_retry_call(func, *args):
    """Fixes OSErrors and IOErrors

    From: http://code.google.com/p/seascope/source/detail?spec=svn8dbe5e23d41db673727ce90fd338e9a43f8877e8&name=8dbe5e23d41d&r=8dbe5e23d41db673727ce90fd338e9a43f8877e8
    IOError added
    """
    while True:
        try:
            return func(*args)
        except (OSError, IOError), e: # pragma: no cover
            if e.errno == errno.EINTR:
                continue
            raise


def _add_tool(path):
    # first create classes
    tool_name = os.path.basename(path)
    if isinstance(tool_name, unicode):
        tool_name = tool_name.encode('utf-8')
    if not tool_name.endswith(SUFFIX): # pragma: no cover
        return
    (tool_name, _) = os.path.splitext(tool_name)

    if tool_name in cl_tools: # pragma: no cover
        debug.critical("Package CLTools already added: '%s'" % tool_name)
    try:
        conf = json.load(open(path))
    except ValueError as exc: # pragma: no cover
        debug.critical("Package CLTools could not parse '%s'" % path, exc)
        return

    def compute(self):
        """ 1. read inputs
            2. call with inputs
            3. set outputs
        """
        # add all arguments as an unordered list
        args = [self.conf['command']]
        file_std = 'options' in self.conf and 'std_using_files' in self.conf['options']
        fail_with_cmd = 'options' in self.conf and 'fail_with_cmd' in self.conf['options']
        setOutput = [] # (name, File) - set File contents as output for name
        open_files = []
        stdin = None
        kwargs = {}
        for type, name, klass, options in self.conf['args']:
            type = type.lower()
            klass = klass.lower()
            if "constant" == type:
                flag = 'flag' in options and options['flag']
                if flag:
                    args.append(flag)
                if name:
                    # if flag==name we assume user tried to name a constant
                    if not name == flag:
                        args.append('%s%s' % (options.get('prefix', ''), name))
            elif "input" == type:
                # handle multiple inputs
                values = self.force_get_input_list(name)
                if values and 'list' == klass:
                    values = values[0]
                    klass = options['type'].lower() \
                      if 'type' in options else 'string'
                for value in values:
                    if 'flag' == klass:
                        if not value:
                            continue
                        if 'flag' in options and options['flag']:
                            value = options['flag']
                        else:
                            # use name as flag
                            value = name
                    elif klass in ('file', 'directory', 'path'):
                        value = value.name
                    # check for flag and append file name
                    if not 'flag' == klass and 'flag' in options:
                        args.append(options['flag'])
                    value = '%s%s' % (options.get('prefix', ''),
                                      value)
                    args.append(value)
            elif "output" == type:
                # output must be a filename but we may convert the result to a string
                # create new file
                file = self.interpreter.filePool.create_file(
                        suffix=options.get('suffix', DEFAULTFILESUFFIX))
                fname = file.name
                if 'prefix' in options:
                    fname = options['prefix'] + fname
                if 'flag' in options:
                    args.append(options['flag'])
                args.append(fname)
                if "file" == klass:
                    self.set_output(name, file)
                elif "string" == klass:
                    setOutput.append((name, file))
                else:
                    raise ValueError
            elif "inputoutput" == type:
                # handle single file that is both input and output
                value = self.get_input(name)

                # create copy of infile to operate on
                outfile = self.interpreter.filePool.create_file(
                        suffix=options.get('suffix', DEFAULTFILESUFFIX))
                try:
                    shutil.copyfile(value.name, outfile.name)
                except IOError, e: # pragma: no cover
                    raise ModuleError(self,
                                      "Error copying file '%s': %s" %
                                      (value.name, debug.format_exception(e)))
                value = '%s%s' % (options.get('prefix', ''), outfile.name)
                # check for flag and append file name
                if 'flag' in options:
                    args.append(options['flag'])
                args.append(value)
                self.set_output(name, outfile)
        if "stdin" in self.conf:
            name, type, options = self.conf["stdin"]
            type = type.lower()
            if self.has_input(name):
                value = self.get_input(name)
                if "file" == type:
                    if file_std:
                        f = open(value.name, 'rb')
                    else:
                        f = open(value.name, 'rb')
                        stdin = f.read()
                        f.close()
                elif "string" == type:
                    if file_std:
                        file = self.interpreter.filePool.create_file()
                        f = open(file.name, 'wb')
                        f.write(value)
                        f.close()
                        f = open(file.name, 'rb')
                    else:
                        stdin = value
                else: # pragma: no cover
                    raise ValueError
                if file_std:
                    open_files.append(f)
                    kwargs['stdin'] = f.fileno()
                else:
                    kwargs['stdin'] = subprocess.PIPE
        if "stdout" in self.conf:
            if file_std:
                name, type, options = self.conf["stdout"]
                type = type.lower()
                file = self.interpreter.filePool.create_file(
                        suffix=DEFAULTFILESUFFIX)
                if "file" == type:
                    self.set_output(name, file)
                elif "string" == type:
                    setOutput.append((name, file))
                else: # pragma: no cover
                    raise ValueError
                f = open(file.name, 'wb')
                open_files.append(f)
                kwargs['stdout'] = f.fileno()
            else:
                kwargs['stdout'] = subprocess.PIPE
        if "stderr" in self.conf:
            if file_std:
                name, type, options = self.conf["stderr"]
                type = type.lower()
                file = self.interpreter.filePool.create_file(
                        suffix=DEFAULTFILESUFFIX)
                if "file" == type:
                    self.set_output(name, file)
                elif "string" == type:
                    setOutput.append((name, file))
                else: # pragma: no cover
                    raise ValueError
                f = open(file.name, 'wb')
                open_files.append(f)
                kwargs['stderr'] = f.fileno()
            else:
                kwargs['stderr'] = subprocess.PIPE

        if fail_with_cmd:
            return_code = 0
        else:
            return_code = self.conf.get('return_code', None)

        env = {}
        # 0. add defaults
        # 1. add from configuration
        # 2. add from module env
        # 3. add from env port
        if configuration.check('env'):
            try:
                for var in configuration.env.split(";"):
                    key, value = var.split('=')
                    key = key.strip()
                    value = value.strip()
                    if key:
                        env[key] = value
            except Exception, e: # pragma: no cover
                raise ModuleError(self,
                                  "Error parsing configuration env: %s" % (
                                  debug.format_exception(e)))

        if 'options' in self.conf and 'env' in self.conf['options']:
            try:
                for var in self.conf['options']['env'].split(";"):
                    key, value = var.split('=')
                    key = key.strip()
                    value = value.strip()
                    if key:
                        env[key] = value
            except Exception, e: # pragma: no cover
                raise ModuleError(self,
                                  "Error parsing module env: %s" % (
                                  debug.format_exception(e)))
            
        if 'options' in self.conf and 'env_port' in self.conf['options']:
            for e in self.force_get_input_list('env'):
                try:
                    for var in e.split(';'):
                        if not var:
                            continue
                        key, value = var.split('=')
                        key = key.strip()
                        value = value.strip()
                        if key:
                            env[key] = value
                except Exception, e: # pragma: no cover
                    raise ModuleError(self,
                                      "Error parsing env port: %s" % (
                                      debug.format_exception(e)))

        if env:
            kwargs['env'] = dict(os.environ)
            kwargs['env'].update(env)
            # write to execution provenance
            env = ';'.join(['%s=%s'%(k,v) for k,v in env.iteritems()])
            self.annotate({'execution_env': env})

        if 'dir' in self.conf:
            kwargs['cwd'] = self.conf['dir']

        process = subprocess.Popen(args, **kwargs)
        if file_std:
            process.wait()
        else:
            #if stdin:
            #    print "stdin:", len(stdin), stdin[:30]
            stdout, stderr = _eintr_retry_call(process.communicate, stdin)
            #stdout, stderr = process.communicate(stdin)
            #if stdout:
            #    print "stdout:", len(stdout), stdout[:30]
            #if stderr:
            #    print "stderr:", len(stderr), stderr[:30]

        if return_code is not None:
            if process.returncode != return_code:
                raise ModuleError(self, "Command returned %d (!= %d)" % (
                                  process.returncode, return_code))
        self.set_output('return_code', process.returncode)

        for f in open_files:
            f.close()

        for name, file in setOutput:
            f = open(file.name, 'rb')
            self.set_output(name, f.read())
            f.close()

        if not file_std:
            if "stdout" in self.conf:
                name, type, options = self.conf["stdout"]
                type = type.lower()
                if "file" == type:
                    file = self.interpreter.filePool.create_file(
                            suffix=DEFAULTFILESUFFIX)
                    f = open(file.name, 'wb')
                    f.write(stdout)
                    f.close()
                    self.set_output(name, file)
                elif "string" == type:
                    self.set_output(name, stdout)
                else: # pragma: no cover
                    raise ValueError
            if "stderr" in self.conf:
                name, type, options = self.conf["stderr"]
                type = type.lower()
                if "file" == type:
                    file = self.interpreter.filePool.create_file(
                            suffix=DEFAULTFILESUFFIX)
                    f = open(file.name, 'wb')
                    f.write(stderr)
                    f.close()
                    self.set_output(name, file)
                elif "string" == type:
                    self.set_output(name, stderr)
                else: # pragma: no cover
                    raise ValueError


    # create docstring
    d = """This module is a wrapper for the command line tool '%s'""" % \
        conf['command']
    # create module
    M = new_module(CLTools, tool_name, {"compute": compute,
                                        "conf": conf,
                                        "tool_name": tool_name,
                                        "__doc__": d})
    reg = vistrails.core.modules.module_registry.get_module_registry()
    reg.add_module(M, package=identifiers.identifier,
                   package_version=identifiers.version)

    def to_vt_type(s):
        # add recognized types here - default is String
        return '(basic:%s)' % \
          {'file':'File', 'path':'Path', 'directory': 'Directory',
           'flag':'Boolean', 'list':'List',
           'float':'Float','integer':'Integer'
          }.get(s.lower(), 'String')
    # add module ports
    if 'stdin' in conf:
        name, type, options = conf['stdin']
        optional = 'required' not in options
        reg.add_input_port(M, name, to_vt_type(type), optional=optional)
    if 'stdout' in conf:
        name, type, options = conf['stdout']
        optional = 'required' not in options
        reg.add_output_port(M, name, to_vt_type(type), optional=optional)
    if 'stderr' in conf:
        name, type, options = conf['stderr']
        optional = 'required' not in options
        reg.add_output_port(M, name, to_vt_type(type), optional=optional)
    if 'options' in conf and 'env_port' in conf['options']:
        reg.add_input_port(M, 'env', to_vt_type('string'))
    for type, name, klass, options in conf['args']:
        optional = 'required' not in options
        if 'input' == type.lower():
            reg.add_input_port(M, name, to_vt_type(klass), optional=optional)
        elif 'output' == type.lower():
            reg.add_output_port(M, name, to_vt_type(klass), optional=optional)
        elif 'inputoutput' == type.lower():
            reg.add_input_port(M, name, to_vt_type('file'), optional=optional)
            reg.add_output_port(M, name, to_vt_type('file'), optional=optional)
    reg.add_output_port(M, 'return_code', to_vt_type('integer'))
    cl_tools[tool_name] = M


def add_tool(path):
    try:
        _add_tool(path)
    except Exception as exc:  # pragma: no cover
        import traceback
        debug.critical("Package CLTools failed to create module "
           "from '%s': %s" % (path, exc),
           traceback.format_exc())


def initialize(*args, **keywords):
    reload_scripts(initial=True)


def remove_all_scripts():
    reg = vistrails.core.modules.module_registry.get_module_registry()
    for tool_name in cl_tools.keys():
        del cl_tools[tool_name]
        reg.delete_module(identifiers.identifier, tool_name)

def reload_scripts(initial=False, name=None):
    reg = vistrails.core.modules.module_registry.get_module_registry()
    if not initial:
        if name is None:
            remove_all_scripts()
        else:
            del cl_tools[name]
            reg.delete_module(identifiers.identifier, name)

    if "CLTools" == identifiers.name:
        # this is the original package
        location = os.path.join(vistrails.core.system.current_dot_vistrails(),
                                "CLTools")
        # make sure dir exist
        if not os.path.isdir(location): # pragma: no cover # pragma: no branch
            try:
                debug.log("Creating CLTools directory...")
                os.mkdir(location)
            except Exception, e:
                debug.critical("Could not create CLTools directory. Make "
                               "sure '%s' does not exist and parent directory "
                               "is writable" % location,
                               e)
                sys.exit(1)
    else:  # pragma: no cover
        # this is a standalone package so modules are placed in this directory
        location = os.path.dirname(__file__)

    if initial:
        reg.add_module(CLTools, abstract=True)
    if name is None:
        for path in os.listdir(location):
            if path.endswith(SUFFIX):  # pragma: no branch
                add_tool(os.path.join(location, path))
    else:
        path = os.path.join(location, name + SUFFIX)
        if os.path.exists(path):
            add_tool(path)

    if not initial:
        from vistrails.core.interpreter.cached import CachedInterpreter
        CachedInterpreter.clear_package(identifiers.identifier)

        from vistrails.gui.vistrails_window import _app
        _app.invalidate_pipelines()


wizards_list = []

def menu_items():
    """menu_items() -> tuple of (str,function)
    It returns a list of pairs containing text for the menu and a
    callback function that will be executed when that menu item is selected.
    
    """
    try:
        from wizard import QCLToolsWizardWindow
    except Exception, e: # pragma: no cover
        if "CLTools" == identifiers.name:
            debug.unexpected_exception(e)
            raise
        else:
            return
    lst = []
    if "CLTools" == identifiers.name: # pragma: no branch
        def open_wizard():
            window = QCLToolsWizardWindow(reload_scripts=reload_scripts)
            wizards_list.append(window)
            window.show()
        lst.append(("Open Wizard", open_wizard))
    lst.append(("Reload All Scripts", reload_scripts))
    return tuple(lst)


def finalize():
    pass


def contextMenuName(name):
    if "CLTools" == name:
        return "Reload All Scripts"
    else:
        return "Reload Script"


def callContextMenu(name):
    if "CLTools" == name:
        reload_scripts()
    else:
        reload_scripts(name=name)


###############################################################################

import unittest
from vistrails.tests.utils import execute, intercept_results


class TestCLTools(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # first make sure CLTools is loaded
        pm = get_package_manager()
        if 'CLTools' not in pm._package_list: # pragma: no cover # pragma: no branch
            pm.late_enable_package('CLTools')
        remove_all_scripts()
        cls.testdir = os.path.join(packages_directory(), 'CLTools', 'test_files')
        cls._tools = {}
        for name in os.listdir(cls.testdir):
            if not name.endswith(SUFFIX):
                continue
            _add_tool(os.path.join(cls.testdir, name))
            toolname = os.path.splitext(name)[0]
            cls._tools[toolname] = cl_tools[toolname]
        cls._old_dir = os.getcwd()
        os.chdir(vistrails_root_directory())

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls._old_dir)
        reload_scripts()

    def do_the_test(self, toolname):
        with intercept_results(
                self._tools[toolname],
                'return_code', 'f_out', 'stdout') as (
                return_code, f_out, stdout):
            self.assertFalse(execute([
                    (toolname, 'org.vistrails.vistrails.cltools', [
                        ('f_in', [('File', self.testdir + '/test_1.cltest')]),
                        ('chars', [('List', '["a", "b", "c"]')]),
                        ('false', [('Boolean', 'False')]),
                        ('true', [('Boolean', 'True')]),
                        ('nb', [('Integer', '42')]),
                        ('stdin', [('String', 'some line\nignored')]),
                    ]),
                ]))
        self.assertEqual(return_code, [0])
        self.assertEqual(f_out, ['ok\nmessage received'])
        self.assertEqual(stdout, ['program output here'])

    def test_with_pipes(self):
        """Without std_using_files: use pipes instead of files.
        """
        self.do_the_test('intern_cltools_1')

    def test_with_files(self):
        """With std_using_files: use files instead of pipes.
        """
        self.do_the_test('intern_cltools_2')



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from vistrails.core.bundles.pyimport import py_import
from vistrails.core import debug
from vistrails.core.modules.vistrails_module import Module, ModuleError

from ..common import Table


def get_xlwt():
    try:
        return py_import('xlwt', {
                             'pip': 'xlwt',
                             'linux-debian': 'python-xlwt',
                             'linux-ubuntu': 'python-xlwt',
                             'linux-fedora': 'python-xlwt'},
                         True)
    except ImportError: # pragma: no cover
        return None


class WriteExcelSpreadsheet(Module):
    """Writes a table to an Excel spreadsheet file.
    """
    _input_ports = [('table', Table)]
    _output_ports = [('file', '(org.vistrails.vistrails.basic:File)')]

    def compute(self):
        table = self.get_input('table')
        rows = table.rows

        xlwt = get_xlwt()
        if xlwt is None: # pragma: no cover
            raise ModuleError(self, "xlwt is not available")

        workbook = xlwt.Workbook()
        sheet = workbook.add_sheet('Sheet1')

        fileobj = self.interpreter.filePool.create_file(suffix='.xls')
        fname = fileobj.name

        for c in xrange(table.columns):
            column = table.get_column(c)
            for r, e in enumerate(column):
                sheet.write(r, c, e)
            if r+1 != rows: # pragma: no cover
                debug.warning("WriteExcelSpreadsheet wrote %d lines instead "
                              "of expected %d" % (r, rows))

        workbook.save(fname)
        self.set_output('file', fileobj)


_modules = [WriteExcelSpreadsheet]



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import sys
import os.path
import shutil
import hashlib
import vistrails.core.system
import vistrails.core.modules.module_registry
import vistrails.core.modules.basic_modules
import traceback
from vistrails.core.packagemanager import get_package_manager
from vistrails.core.modules.package import Package
from vistrails.core.modules.vistrails_module import Module, ModuleError, new_module
from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler
from vistrails.core import debug

import suds


package_cache = None

webServicesDict = {}

def toSignature(s):
    """ transform wsdl address to valid package signature """
    # cannot have ':' in package name
    return 'SUDS#' + s.replace(':', '$')

def toAddress(s):
    """ transform package signature to valid wsdl address """
    # look up address for a specific signature
    return s[5:].replace('$', ':')

###############################################################################

wsdlSchemas = ['http://www.w3.org/2001/XMLSchema',
              'http://xml.apache.org/xml-soap',
              'http://schemas.xmlsoap.org/soap/encoding/']
#Dictionary of primitive types
#list extracted from http://www.w3.org/2001/XMLSchema
wsdlTypesDict = { 'string' : vistrails.core.modules.basic_modules.String,
                'float': vistrails.core.modules.basic_modules.Float,
                'decimal': vistrails.core.modules.basic_modules.Float,
                'double': vistrails.core.modules.basic_modules.Float,
                'boolean': vistrails.core.modules.basic_modules.Boolean,
                'anyType': vistrails.core.modules.basic_modules.Variant,
                'anyURI': vistrails.core.modules.basic_modules.String,
                'DataHandler':vistrails.core.modules.basic_modules.String,
                'dateTime': vistrails.core.modules.basic_modules.String,
                'duration': vistrails.core.modules.basic_modules.String,
                'time': vistrails.core.modules.basic_modules.String,
                'date': vistrails.core.modules.basic_modules.String,
                'gYearMonth': vistrails.core.modules.basic_modules.String,
                'gYear': vistrails.core.modules.basic_modules.String,
                'gMonthDay': vistrails.core.modules.basic_modules.String,
                'gDay': vistrails.core.modules.basic_modules.String,
                'gMonth': vistrails.core.modules.basic_modules.String,
                'hexBinary': vistrails.core.modules.basic_modules.String,
                'base64': vistrails.core.modules.basic_modules.String,
                'base64Binary': vistrails.core.modules.basic_modules.String,
                'QName': vistrails.core.modules.basic_modules.String,
                'normalizedString': vistrails.core.modules.basic_modules.String,
                'token': vistrails.core.modules.basic_modules.String,
                'language': vistrails.core.modules.basic_modules.String,
                'IDREFS': vistrails.core.modules.basic_modules.String,
                'ENTITIES': vistrails.core.modules.basic_modules.String,
                'NMTOKEN': vistrails.core.modules.basic_modules.String,
                'NMTOKENS': vistrails.core.modules.basic_modules.String,
                'Name': vistrails.core.modules.basic_modules.String,
                'NCName': vistrails.core.modules.basic_modules.String,
                'ID': vistrails.core.modules.basic_modules.String,
                'IDREF': vistrails.core.modules.basic_modules.String,
                'ENTITY': vistrails.core.modules.basic_modules.String,
                'integer' : vistrails.core.modules.basic_modules.Integer,
                'nonPositiveInteger' : vistrails.core.modules.basic_modules.Integer,
                'negativeInteger' : vistrails.core.modules.basic_modules.Integer,
                'long' : vistrails.core.modules.basic_modules.Integer,
                'int' : vistrails.core.modules.basic_modules.Integer,
                'short' : vistrails.core.modules.basic_modules.Integer,
                'byte' : vistrails.core.modules.basic_modules.Integer,
                'nonNegativeInteger' : vistrails.core.modules.basic_modules.Integer,
                'unsignedLong' : vistrails.core.modules.basic_modules.Integer,
                'unsignedInt' : vistrails.core.modules.basic_modules.Integer,
                'unsignedShort' : vistrails.core.modules.basic_modules.Integer,
                'unsignedByte' : vistrails.core.modules.basic_modules.Integer,
                'positiveInteger' : vistrails.core.modules.basic_modules.Integer,
                'Array': vistrails.core.modules.basic_modules.List,

                # soapenc types
                'arrayCoordinate': vistrails.core.modules.basic_modules.String,
                'Struct': vistrails.core.modules.basic_modules.String,
                'NOTATION': vistrails.core.modules.basic_modules.String                
                }

def initialize(*args, **keywords):
#    import core.packagemanager
    global webServicesDict
    global package_cache

    #Create a directory for the SUDSWebServices package
    location = os.path.join(vistrails.core.system.current_dot_vistrails(),
                            "SUDSWebServices")
    if not os.path.isdir(location):
        try:
            debug.log("Creating SUDS cache directory...")
            os.mkdir(location)
        except OSError, e:
            debug.critical(
"""Could not create SUDS cache directory. Make sure
'%s' does not exist and parent directory is writable""" % location, e)
            sys.exit(1)
    # the number of days to cache wsdl files
    days = 1
    if configuration.check("cache_days"):
        days = configuration.cache_days
    suds.client.ObjectCache.protocol = 0 # windows needs this
    package_cache = suds.client.ObjectCache(location, days=days)

    #reg = core.modules.module_registry.get_module_registry()
    #reg.add_module(SUDSWebService, abstract=True)

    wsdlList = []
    if configuration.check('wsdlList'):
        wsdlList = configuration.wsdlList.split(";")
    else:
        configuration.wsdlList = ''
    for wsdl in wsdlList:
        if not wsdl:
            break
        if not wsdl.startswith('http://'):
            wsdl = 'http://' + wsdl
        if wsdl in webServicesDict:
            debug.warning('Duplicate WSDL entry: '+wsdl)
            continue
        s = Service(wsdl)
        if s.service:
            webServicesDict[wsdl] = s
        
def finalize():
    # unload service packages
    reg = vistrails.core.modules.module_registry.get_module_registry()
    for s in webServicesDict.itervalues():
        if s.package:
            reg.remove_package(s.package)

class WSMethod(object):
    """ A WSDL method
    """
    def __init__(self, qname=('','')):
        self.qname = qname
        "name: (typename, ns)"
        self.inputs = {}
        self.outputs = {}

class WSElement(object):
    """ A part of a WSDL type
    """
    def __init__(self, name='', type=('',''), optional=False, min=0,
                 max='unbounded', enum=False):
        self.name = name
        self.type = type
        self.optional = optional
        self.min = min
        self.max = max
        self.enum = enum

class WSType(object):
    """ A WSDL type definition
    """
    def __init__(self, qname=('',''), enum=False):
        self.qname = qname
        self.enum = enum
        "name: WSElement"
        self.parts = {}
 
class Service(object):
    def __init__(self, address):
        """ Process WSDL and add all Types and Methods
        """
        self.address = address
        self.signature = toSignature(self.address)
        self.wsdlHash = '-1'
        self.modules = []
        self.package = None
        debug.log("Installing Web Service from WSDL: %s"% address)

        options = dict(cachingpolicy=1, cache=package_cache)
        
        proxy_types = ['http']
        for t in proxy_types:
            key = 'proxy_%s'%t
            if configuration.check(key):
                proxy = getattr(configuration, key)
                debug.log("Using proxy: %s" % proxy)
                if len(proxy):
                    options['proxy'] = {t:proxy}
        try:
            self.service = suds.client.Client(address, **options)
            self.backUpCache()
        except Exception:
            self.service = None
            # We may be offline and the cache may have expired,
            # try to use backup
            if self.restoreFromBackup():
                try:
                    self.service = suds.client.Client(address, **options)
                except Exception:
                    self.service = None
                    debug.critical("Could not load WSDL: %s" % address,
                                   traceback.format_exc())
            else:
                debug.critical("Could not load WSDL: %s" % address,
                               traceback.format_exc())
        if self.service:
            try:
                self.createPackage()
                self.setTypes()
                self.setMethods()
                self.createTypeClasses()
                self.createMethodClasses()
            except Exception:
                debug.critical("Could not create Web Service: %s" % address,
                               traceback.format_exc())
                self.service = None
        if self.wsdlHash == '-1':
            # create empty package so that it can be reloaded/deleted
            self.createFailedPackage()
        
    def makeDictType(self, obj):
        """ Create recursive dict from SUDS object
            ignore attributes with None and empty dicts and handle lists
        """
        if not hasattr(obj,'__keylist__'):
            return obj
        if isinstance(obj, list):
            p = [self.makeDictType(i) for i in obj if i is not None]
            return [i for i in p if i is not None and i != {}]
        d = {}
        for o in obj.__keylist__:
            e = getattr(obj, o)
            if e is not None:
                p = self.makeDictType(e)
                if p is not None and p != {}:
                    d[str(o)] = self.makeDictType(e)
        return d
            
    def backUpCache(self):
        """ We back up the cache for situations where the cache is expired and
            the web service can not be reached. This allows working offline.
        """ 
        cached = os.path.join(package_cache.location,
                              "suds-%s-wsdl.px" % abs(hash(self.address)))
        backup = os.path.join(package_cache.location,
                              "suds-%s-wsdl.px" % hashlib.md5(self.address
                                                              ).hexdigest())
        if os.path.exists(cached):
            shutil.copyfile(cached, backup)

    def restoreFromBackup(self):
        """ Restore from backup into cache with reset cache time""" 
        cached = os.path.join(package_cache.location,
                              "suds-%s-wsdl.px" % abs(hash(self.address)))
        backup = os.path.join(package_cache.location,
                              "suds-%s-wsdl.px" % hashlib.md5(self.address
                                                              ).hexdigest())
        if not os.path.exists(cached) and os.path.exists(backup):
            shutil.copyfile(backup, cached)
            return True
        return False

    def createPackage(self):
        reg = vistrails.core.modules.module_registry.get_module_registry()
        if self.signature in reg.packages:
            reg.remove_package(reg.packages[self.signature])

        # create a document hash integer from the cached sax tree
        # "name" is what suds use as the cache key
        name = '%s-%s' % (abs(hash(self.address)), "wsdl")
        wsdl = package_cache.get(name)
        if not wsdl:
            debug.critical("File not found in SUDS cache: '%s'" % name)
            self.wsdlHash = '0'
            return
        self.wsdlHash = str(int(hashlib.md5(str(wsdl.root)).hexdigest(), 16))

        package_id = reg.idScope.getNewId(Package.vtType)
        package = Package(id=package_id,
                          load_configuration=False,
                          name="SUDS#" + self.address,
                          identifier=self.signature,
                          version=self.wsdlHash,
                          )
        suds_package = reg.get_package_by_name(identifier)
        package._module = suds_package.module
        package._init_module = suds_package.init_module
        
        self.package = package
        reg.add_package(package)
        reg.signals.emit_new_package(self.signature)

        self.module = new_module(Module, str(self.signature))
        reg.add_module(self.module, **{'package':self.signature,
                                       'package_version':self.wsdlHash,
                                       'abstract':True})

    def createFailedPackage(self):
        """ Failed package is created so that the user can remove
        it manually using package submenu """
        pm = get_package_manager()
        if pm.has_package(self.signature):
            # do nothing
            return
        reg = vistrails.core.modules.module_registry.get_module_registry()

        # create a document hash integer from the cached sax tree
        # "name" is what suds use as the cache key
        name = '%s-%s' % (abs(hash(self.address)), "wsdl")
        self.wsdlHash = '0'

        package_id = reg.idScope.getNewId(Package.vtType)
        package = Package(id=package_id,
                          load_configuration=False,
                          name="SUDS#" + self.address,
                          identifier=self.signature,
                          version=self.wsdlHash,
                          )
        suds_package = reg.get_package_by_name(identifier)
        package._module = suds_package.module
        package._init_module = suds_package.init_module
        self.package = package
        reg.add_package(package)
        reg.signals.emit_new_package(self.signature)
        self.module = new_module(Module, str(self.signature))
        reg.add_module(self.module, **{'package':self.signature,
                                       'package_version':self.wsdlHash,
                                       'abstract':True})
        self.service = -1

    def setTypes(self):
        """ Return dict containing all exposed types in the service
        """
        self.wstypes = {}
        for sd in self.service.sd:
            for type in sd.types:
                type = type[0]
                # ignore existing types and arrays
                if type.qname[1] in wsdlSchemas or type.qname[0].startswith('Array'):
                    continue
                qname = type.qname
                wstype = WSType(qname, type.enum())
                self.wstypes[qname] = wstype
                # handle enums by adding a "value" port with the possible enums
                if wstype.enum:
                    values = [i[0].name for i in type.children()]
                    wstype.parts["value"] = WSElement("value",
                        ('string', "http://www.w3.org/2001/XMLSchema"),
                        False, 1, 1, values)
                    continue
                # add elements
                for e in type.children():
                    e = e[0]
                    if e.type is None:
                        pass # probably empty type so just ignore
                    elif e.type[0].startswith('Array'):
                        qtype = ('Array', 'http://www.w3.org/2001/XMLSchema')
                        wstype.parts[e.name] = WSElement(e.name, qtype,
                                                   e.optional(), e.min, e.max)
                    else:
                        wstype.parts[e.name] = WSElement(e.name, e.type,
                                                   e.optional(), e.min, e.max)

    def setMethods(self):
        """ Return dict with methods in service
        """
        def isKnownType(qtype):
            return not qtype or str(qtype[1]) in wsdlSchemas or \
            qtype[0].startswith('Array') or qtype in self.wstypes

        def getChildTypes(type):
            schema = self.service.wsdl.schema
            if type in schema.elements:
                if schema.elements[type].type is not None:
                    return [(schema.elements[type].name, schema.elements[type].type)]
                else:
                    return [(c[0].name, c[0].type) for c in schema.elements[type].children()]
            elif type in schema.types:
                return [(c[0].name, c[0].type) for c in schema.types[type].children()]
            debug.critical("Could not resolve type in WSDL:" + str(type))
            return []

        def setMethod(m):
            # create new method
            wsm = WSMethod((m[0], self.address))
            self.wsmethods[wsm.qname] = wsm
            # process inputs
            for param in m[1]:
                name = param[0]
                qtype = (param[1].resolve().name, param[1].resolve().namespace()[1])
                qtype = (str(qtype[0]), str(qtype[1]))
                if qtype is None:
                    qtype = param[1].element
                    qtype = (str(qtype[0]), str(qtype[1]))
                if qtype is not None:
                    if qtype[0].startswith('Array'):
                        qtype = ('Array', 'http://www.w3.org/2001/XMLSchema')
                    wsm.inputs[name] = qtype
            # process outputs - need to look recursively in wsdl
            method = self.service.service.__getattr__(m[0])
            for output in method.method.soap.output.body.parts:
                # get schema definition of method
                if output.type:
                    qtype = output.type
                else:
                    element = self.service.wsdl.schema.elements[output.element]
                    qtype = element.type if element.type else output.element
                # check that it exists
                types = [(output.name, qtype)]
                while qtype is not None and not isKnownType(qtype):
                    types = getChildTypes(qtype)
                    if len(types):
                        qtype = types[0][1]
                    else:
                        break
                for name, qtype in types:
                    if qtype is not None:
                        if qtype[0].startswith('Array'):
                            qtype = ('Array', 'http://www.w3.org/2001/XMLSchema')
                        wsm.outputs[name] = qtype
        # loop through services, ports, and methods
        self.wsmethods = {}
        for sd in self.service.sd:
            for port in sd.ports:
                for m in port[1]:
                    setMethod(m)

    def createTypeClasses(self):
        # first create classes
        reg = vistrails.core.modules.module_registry.get_module_registry()
        self.typeClasses = {}
        for t in self.wstypes.itervalues():
            def compute(self):
                """ 1. use type input as object or create new
                    2. add other inputs to obj
                    3. set obj and parts as outputs
                """
                if self.wstype.enum:
                    # only makes sure the enum is one of the valid values
                    p = self.wstype.parts['value']
                    if self.has_input(p.name):
                        obj = self.get_input(p.name)
                    else:
                        obj = p.enum[0] if len(p.enum) else ''
                    if self.has_input('value'):
                        obj = self.get_input('value')
                    if obj not in p.enum:
                        raise ModuleError(self,
                                 "'%s' is not one of the valid enums: %s" %
                                 (obj, str(p.enum)) )
                    self.set_output(self.wstype.qname[0], obj)
                    self.set_output('value', obj)
                    return
                if self.has_input(self.wstype.qname[0]):
                    obj = self.get_input(self.wstype.qname[0])
                else:
                    obj = {}
                    s = "{%s}%s"%(self.wstype.qname[1],self.wstype.qname[0])
                    try:
                        obj = self.service.service.factory.create(s)
                    except (suds.TypeNotFound, suds.BuildError):
                        raise ModuleError("Type not found: %s" % s)
                for part in self.wstype.parts.itervalues():
                    # 
                    if obj.__class__.__name__ == 'UberClass':
                        # UberClass is a placeholder and its value is assumed
                        # to be the correct attribute value
                        if len(self.wstype.parts) == 1:
                            setattr(obj, part.name, obj.value)
                        else:
                            # update each attribute
                            if hasattr(obj.value, part.name):
                                setattr(obj, part.name, getattr(obj.value, part.name))
                    if self.has_input(part.name):
                        p = self.get_input(part.name)
                        if hasattr(obj, part.name):
                            setattr(obj, part.name, p)
                        else:
                            # do it anyway - assume attribute missing in template
                            setattr(obj, part.name, p)
                    if hasattr(obj, part.name):
                        # 
                        res = getattr(obj, part.name)
                        self.set_output(part.name, res)
                self.set_output(self.wstype.qname[0], obj)

            # create docstring
            parts = ", ".join([i.type[0]+' '+i.name for i in t.parts.itervalues()])
            d = """This module was created using a wrapper for SUDS (fedorahosted.org/suds/)
from the WSDL spec at:
   %s
It is a WSDL type with signature:
   %s(%s)"""%(self.address, t.qname[0], parts)
            M = new_module(self.module, str(t.qname[0]),{"compute":compute,
                                                         "wstype":t,
                                                         "service":self,
                                                         "__doc__":d})
            self.typeClasses[t.qname] = M
            reg.add_module(M, **{'namespace':'Types',
                                 'package':self.signature,
                                 'package_version':self.wsdlHash})

        # then add ports
        for t in self.wstypes:
            wstype = self.wstypes[t]
            # get type module
            if t[1] in wsdlSchemas:
                c = wsdlTypesDict[t[0]]
            elif t in self.typeClasses:
                c = self.typeClasses[t]
            else:
                debug.critical("Cannot find module for type: " + str(t))
                continue
            # add self ports
            reg.add_input_port(self.typeClasses[t], t[0], c)
            reg.add_output_port(self.typeClasses[t], t[0], c)
            for p in wstype.parts:
                part = wstype.parts[p]
                # get type module
                ptype = part.type
                if part.max is not None and (part.max == 'unbounded' or int(part.max)>1):
                    # it can be multiple objects which means we need to make it a list
                    c = vistrails.core.modules.basic_modules.List
                elif ptype[1] in wsdlSchemas:
                    c = wsdlTypesDict[ptype[0]]
                elif ptype in self.typeClasses:
                    c = self.typeClasses[ptype]
                else:
                    debug.critical("Cannot find module for type: " + str(ptype))
                    continue
                # add as both input and output port
                reg.add_input_port(self.typeClasses[t], p, c,
                                   optional=part.optional)
                reg.add_output_port(self.typeClasses[t], p, c,
                                    optional=part.optional)

    def createMethodClasses(self):
        # register modules
        reg = vistrails.core.modules.module_registry.get_module_registry()
        self.methodClasses = {}
        for m in self.wsmethods.itervalues():
            def compute(self):
                # create dict of inputs
                cacheable = False
                if self.has_input('cacheable'):
                    cacheable = self.get_input('cacheable')
                self.is_cacheable = lambda *args, **kwargs: cacheable            
                params = {}
                mname = self.wsmethod.qname[0]
                for name in self.wsmethod.inputs:
                    name = str(name)
                    if self.has_input(name):
                        params[name] = self.get_input(name)
                        if params[name].__class__.__name__ == 'UberClass':
                            params[name] = params[name].value
                        params[name] = self.service.makeDictType(params[name])
                try:
                    #import logging
                    #logging.basicConfig(level=logging.INFO)
                    #logging.getLogger('suds.client').setLevel(logging.DEBUG)
                    #print "params:", str(params)[:400]
                    #self.service.service.set_options(retxml = True)
                    #result = getattr(self.service.service.service, mname)(**params)
                    #print "result:", str(result)[:400]
                    #self.service.service.set_options(retxml = False)
                    result = getattr(self.service.service.service, mname)(**params)
                except Exception, e:
                    debug.unexpected_exception(e)
                    raise ModuleError(self, "Error invoking method %s: %s" % (
                            mname, debug.format_exception(e)))
                for name, qtype in self.wsmethod.outputs.iteritems():
                    if isinstance(result, list):
                        # if result is a list just set the output
                        self.set_output(name, result)
                    elif qtype[0] == 'Array':
                        # if result is a type but type is a list try to extract the correct element
                        if len(result.__keylist__):
                            self.set_output(name, getattr(result, result.__keylist__[0]))
                        else:
                            self.set_output(name, result)
                    elif result.__class__.__name__ == 'Text':
                        # only text returned so we assume each output wants all of it
                        self.set_output(name, str(result.trim()))
                    elif result.__class__.__name__ == qtype[0]:
                        # the return value is this type
                        self.set_output(name, result)
                    elif hasattr(result, name):
                        self.set_output(name, getattr(result, name))
                    else:
                        # nothing matches - assume it is an attribute of the correct class
                        class UberClass(object):
                            def __init__(self, value):
                                self.value = value
                        self.set_output(name, UberClass(result))

            # create docstring
            inputs = ", ".join([t[0]+' '+i for i,t in m.inputs.iteritems()])
            outputs = ", ".join([t[0]+' '+o for o,t in m.outputs.iteritems()])
            d = """This module was created using a wrapper for SUDS (fedorahosted.org/suds/)
from the WSDL spec at:
   %s
It is a WSDL method with signature:
   %s(%s)
Outputs:
   (%s)
"""%(self.address, m.qname[0], inputs, outputs)

            M = new_module(self.module, str(m.qname[0]), {"compute":compute,
                                                          "wsmethod":m,
                                                          "service":self,
                                                           "__doc__":d})
            self.methodClasses[m.qname] = M
            reg.add_module(M, **{'namespace':'Methods',
                                 'package':self.signature,
                                 'package_version':self.wsdlHash})
            reg.add_input_port(self.methodClasses[m.qname], 'cacheable',
                               wsdlTypesDict['boolean'], optional=True)

            # add ports
            for p, ptype in m.inputs.iteritems():
                if ptype[1] in wsdlSchemas:
                    c = wsdlTypesDict[ptype[0]]
                elif ptype in self.typeClasses:
                    c = self.typeClasses[ptype]
                else:
                    # use string as default
                    c = wsdlTypesDict['string']
                reg.add_input_port(M, p, c)
            for p, ptype in m.outputs.iteritems():
                if ptype[1] in wsdlSchemas:
                    c = wsdlTypesDict[ptype[0]]
                elif ptype in self.typeClasses:
                    c = self.typeClasses[ptype]
                else:
                    # use string as default
                    c = wsdlTypesDict['string']
                reg.add_output_port(M, p, c)

def load_from_signature(signature):
    """ Load wsdl from signature and return success status """
    wsdl = toAddress(signature)
    wsdlList = []
    if configuration.check('wsdlList'):
        wsdlList = configuration.wsdlList.split(";")
    if not wsdl in wsdlList:
        try:
            service = Service(wsdl)
        except Exception, e:
            debug.unexpected_exception(e)
            return False
        if not service.service:
            return False
        webServicesDict[wsdl] = service
        wsdlList.append(wsdl)
        configuration.wsdlList = ';'.join(wsdlList)
    return True

def handle_module_upgrade_request(controller, module_id, pipeline):
    old_module = pipeline.modules[module_id]
    # first check package
    # v1.0 types:
    if old_module.package == 'edu.utah.sci.vistrails.sudswebservices':
        wsdl = old_module.namespace.split('|')[0]
        namespace = old_module.namespace.split('|')[1]
    else:
        wsdl = toAddress(old_module.package)
        namespace = old_module.namespace
    name = old_module.name

    wsdlList = []
    if configuration.check('wsdlList'):
        wsdlList = configuration.wsdlList.split(";")
    if not wsdl in wsdlList:
        service = Service(wsdl)
        if not service.service:
            return []
        webServicesDict[wsdl] = service
        wsdlList.append(wsdl)
        configuration.wsdlList = ';'.join(wsdlList)

    if old_module.package == 'edu.utah.sci.vistrails.sudswebservices':
        reg = vistrails.core.modules.module_registry.get_module_registry()
        new_descriptor = reg.get_descriptor_by_name(toSignature(wsdl), name,
                                                    namespace)
        if not new_descriptor:
            return []
        try:
            return UpgradeWorkflowHandler.replace_module(controller, pipeline,
                                                    module_id, new_descriptor)
        except Exception, e:
            import traceback
            traceback.print_exc()
            raise

    return UpgradeWorkflowHandler.attempt_automatic_upgrade(controller, 
                                                            pipeline,
                                                            module_id)

def handle_missing_module(controller, module_id, pipeline):
    global webServicesDict
    global package_cache

    def get_wsdl_from_namespace(m_namespace):
        try:
            wsdl = m_namespace.split("|")
            return wsdl[0]
        except Exception, e:
            debug.unexpected_exception(e)
            return None
    
    m = pipeline.modules[module_id]
    if m.package == identifier:
        # v1.0 old style
        m_namespace = m.namespace
        wsdl = get_wsdl_from_namespace(m_namespace)
    else:
        # v1.1 new style
        wsdl = toAddress(m.package)
    
    wsdlList = []
    if configuration.check('wsdlList'):
        wsdlList = configuration.wsdlList.split(";")
    if wsdl in wsdlList:
        # it is already loaded
        return True

    service = Service(wsdl)
    if not service.service:
        return False

    webServicesDict[wsdl] = service
    wsdlList.append(wsdl)
    configuration.wsdlList = ';'.join(wsdlList)
    return True

def load_from_identifier(service_identifier):
    """ Loads a web service from its package identifier """
    global webServicesDict
    global package_cache

    wsdl = toAddress(service_identifier)
    
    wsdlList = []
    if configuration.check('wsdlList'):
        wsdlList = configuration.wsdlList.split(";")
    if wsdl in wsdlList:
        # it is already loaded
        return True

    service = Service(wsdl)
    if not service.service:
        return False

    webServicesDict[wsdl] = service
    wsdlList.append(wsdl)
    configuration.wsdlList = ';'.join(wsdlList)
    return True

def contextMenuName(signature):
    """ Returns a context menu name that depends on the signature
        should return None if no method is available
    """
    if signature == name:
        return "Add Web Service"
    elif signature.startswith('SUDS#'):
        return "Remove this Web Service"
    return None

def saveVistrailFileHook(vistrail, temp_dir):
    """ This is called when a vistrail is loaded
        We should copy all used Web Services in temp_dir to .vistrails
        """

    
    packages = vistrail.get_used_packages()
    # clear old files
    for name in os.listdir(temp_dir):
        if name.endswith("-wsdl.px"):
            os.remove(os.path.join(temp_dir, name))

    for package in packages:
        if package.startswith("SUDS#"):
            address = toAddress(package)
            name = "suds-%s-wsdl.px" % hashlib.md5(address).hexdigest()
            cached = os.path.join(package_cache.location, name)
            vt_cached = os.path.join(temp_dir, name)
            if os.path.exists(cached):
                shutil.copyfile(cached, vt_cached)

def loadVistrailFileHook(vistrail, temp_dir):
    """ This is called when a vistrail is saved
        We should copy all cached Web Services in vistrail to temp_dir
        if they do not exist
    """
    for name in os.listdir(temp_dir):
        src = os.path.join(temp_dir, name)
        dest = os.path.join(package_cache.location, name)
        if name.endswith("-wsdl.px") and not os.path.exists(dest):
            shutil.copyfile(src, dest)
    
def callContextMenu(signature):
    global webServicesDict
    if signature == name:
        from PyQt4 import QtGui 
        wsdl, ret = QtGui.QInputDialog.getText(None, 'Add Web Service',
                                           'Enter the location of the WSDL:')
        wsdl = str(wsdl)
        if not wsdl:
            return
        if not wsdl.startswith('http://'):
            wsdl = 'http://' + wsdl
        if wsdl in webServicesDict:
            debug.critical('WSDL already loaded: '+wsdl)
            return
        s = Service(wsdl)
        if s.service:
            webServicesDict[wsdl] = s
            if configuration.wsdlList:
                configuration.wsdlList += ';' + wsdl
            else:
                configuration.wsdlList = wsdl
    elif signature.startswith('SUDS#'):
        address = toAddress(signature)
        from PyQt4 import QtGui 
        res = QtGui.QMessageBox.question(None,
                           'Remove the following web service from vistrails?',
                           address,
                           buttons=QtGui.QMessageBox.Yes,
                           defaultButton=QtGui.QMessageBox.No)
        if res == QtGui.QMessageBox.Yes:
            # Remove this Web Service
            s = webServicesDict[address]
            del webServicesDict[address]

            reg = vistrails.core.modules.module_registry.get_module_registry()
            reg.remove_package(s.package)
            wsdlList = configuration.wsdlList.split(";")
            wsdlList.remove(address)
            configuration.wsdlList = ';'.join(wsdlList)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

class ConvertVTKToVT(object):
    def __init__(self, in_filename, out_py_file):
        self.in_filename = in_filename
        self.out_py_file = out_py_file
        
        
    def run(self):
        in_file = file(self.in_filename)
        out_file = file(self.out_py_file, 'w')

        does_continue = False
        line_count = 0
        found_starter = False
        for line in in_file:
            if line_count == 0:
                if line.startswith('#!'):
                    found_starter = True
                else:
                    print >> out_file, \
                        "# This file produced automatically by convert_to_vt.py\n# "
            elif line_count == 1 and found_starter:
                print >> out_file, \
                    "# This file produced automatically by convert_to_vt.py\n# "
                found_starter = False
                
            if (line.find('import ') != -1 and line.find('vtk') != -1) or \
                    line.find('VTK_DATA_ROOT = ') != -1 or \
                    does_continue:
                if line.rstrip() == 'import vtk':
                    print >>out_file, 'import sys'
                    print >>out_file, 'sys.path.append' + \
                        '("/vistrails/src/trunk/scripts/gen_vtk_examples")'
                    print >>out_file, 'import vtk_imposter'
                    print >>out_file, 'from colors import *'
                    print >>out_file, 'vtk = vtk_imposter.vtk()'
                    print >>out_file, 'VTK_DATA_ROOT = ""'
                    print >>out_file, ''
                    print >>out_file, '#', line[:-1]
                else:
                    if line.rstrip()[-1] == '\\':
                        does_continue = True
                    else:
                        does_continue = False
                    print >>out_file, '#', line[:-1]
            else:
                out_file.write(line)
            line_count += 1
        print >>out_file, '\nif len(sys.argv) > 1:'
        print >>out_file, '    out_filename = sys.argv[1]'
        print >>out_file, 'else:'
        print >>out_file, '    out_filename = None'
        print >>out_file, 'vtk.to_vt(out_filename)'
        out_file.close()
            

if __name__ == '__main__':
    import sys
    if len(sys.argv) < 3:
        print 'Usage: %s <in_file> <out_py_file>' % sys.argv[0]
        sys.exit(0)
    converter = ConvertVTKToVT(*sys.argv[1:])
    converter.run()
    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from auto_gen import DBWorkflow as _DBWorkflow
from auto_gen import DBAbstractionRef, DBModule, DBGroup
from id_scope import IdScope

import copy

class DBWorkflow(_DBWorkflow):

    def __init__(self, *args, **kwargs):
        _DBWorkflow.__init__(self, *args, **kwargs)
        self.objects = {}
        self.tmp_id = IdScope(1,
                              {DBAbstractionRef.vtType: DBModule.vtType,
                               DBGroup.vtType: DBModule.vtType})

    def __copy__(self):
        return DBWorkflow.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = _DBWorkflow.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = DBWorkflow
        # need to go through and reset the index to the copied objects
        cp.build_index()
        cp.tmp_id = copy.copy(self.tmp_id)
        return cp        

    _vtTypeMap = {'abstractionRef': 'module', 'group': 'module'}

    def build_index(self):
        g = self._vtTypeMap.get
        self.objects = dict(((g(o.vtType, o.vtType), o._db_id), o)
                            for (o,_,_) in self.db_children())

    def add_to_index(self, object):
        obj_type = self._vtTypeMap.get(object.vtType, object.vtType)
        self.objects[(obj_type, object.getPrimaryKey())] = object

    def delete_from_index(self, object):
        obj_type = self._vtTypeMap.get(object.vtType, object.vtType)
        del self.objects[(obj_type, object.getPrimaryKey())]

    def capitalizeOne(self, str):
        return str[0].upper() + str[1:]

    def db_print_objects(self):
        for k,v in self.objects.iteritems():
            print '%s: %s' % (k, v)

    def db_has_object(self, type, id):
        return (type, id) in self.objects

    def db_get_object(self, type, id):
        return self.objects[(type, id)]

    def db_add_object(self, object, parent_obj_type=None,
                      parent_obj_id=None, parent_obj=None):
        if parent_obj is None:
            if parent_obj_type is None or parent_obj_id is None:
                parent_obj = self
            else:
                if parent_obj_type == 'abstractionRef' or \
                        parent_obj_type == 'group':
                    parent_obj_type = 'module'
                try:
                    parent_obj = self.objects[(parent_obj_type, parent_obj_id)]
                except KeyError:
                    msg = "Cannot find object of type '%s' with id '%s'" % \
                        (parent_obj_type, parent_obj_id)
                    raise Exception(msg)
        if object.vtType == 'abstractionRef' or object.vtType == 'group':
            obj_type = 'module'
        else:
            obj_type = object.vtType
        funname = 'db_add_' + obj_type
        obj_copy = copy.copy(object)
        getattr(parent_obj, funname)(obj_copy)
        self.add_to_index(obj_copy)

    def db_change_object(self, old_id, object, parent_obj_type=None, 
                         parent_obj_id=None, parent_obj=None):
        if parent_obj is None:
            if parent_obj_type is None or parent_obj_id is None:
                parent_obj = self
            else:
                if parent_obj_type == 'abstractionRef' or \
                        parent_obj_type == 'group':
                    parent_obj_type = 'module'
                try:
                    parent_obj = self.objects[(parent_obj_type, parent_obj_id)]
                except KeyError:
                    msg = "Cannot find object of type '%s' with id '%s'" % \
                        (parent_obj_type, parent_obj_id)
                    raise Exception(msg)

        self.db_delete_object(old_id, object.vtType, None, None, parent_obj)
        self.db_add_object(object, None, None, parent_obj)

    def db_delete_object(self, obj_id, obj_type, parent_obj_type=None, 
                         parent_obj_id=None, parent_obj=None):
        if parent_obj is None:
            if parent_obj_type is None or parent_obj_id is None:
                parent_obj = self
            else:
                if parent_obj_type == 'abstractionRef' or \
                        parent_obj_type == 'group':
                    parent_obj_type = 'module'
                try:
                    parent_obj = self.objects[(parent_obj_type, parent_obj_id)]
                except KeyError:
                    msg = "Cannot find object of type '%s' with id '%s'" % \
                        (parent_obj_type, parent_obj_id)
                    raise Exception(msg)
        if obj_type == 'abstractionRef' or obj_type == 'group':
            obj_type = 'module'
        funname = 'db_get_' + obj_type
        if hasattr(parent_obj, funname):
            object = getattr(parent_obj, funname)(obj_id)
        else:
            attr_name = 'db_' + obj_type
            object = getattr(parent_obj, attr_name)
        funname = 'db_delete_' + obj_type
        getattr(parent_obj, funname)(object)
        self.delete_from_index(object)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore

from vistrails.gui.base_view import BaseView
from vistrails.gui.paramexplore.pe_table import QParameterExplorationWidget
from vistrails.gui.theme import CurrentTheme
from vistrails.core import debug

class QParamExploreView(QParameterExplorationWidget, BaseView):
    explorationId = 0
    
    def __init__(self, parent=None):
        QParameterExplorationWidget.__init__(self, parent)
        BaseView.__init__(self)

        self.set_title("Explore")
        self.connect(self.table,
                     QtCore.SIGNAL('exploreChange(bool)'),
                     self.exploreChange)

    def set_controller(self, controller):
        if self.controller == controller:
            return
        self.controller = controller
        self.set_exploration()
        
    def updatePipeline(self, pipeline):
        name = self.controller.get_pipeline_name()
        self.set_title("Explore: %s" % name)

    def set_exploration(self, pe=None):
        if not pe:
            pe = self.controller.current_parameter_exploration
        self.setParameterExploration(pe)

    def set_default_layout(self):
        from vistrails.gui.paramexplore.pe_inspector import QParamExploreInspector
        from vistrails.gui.paramexplore.param_view import QParameterView
        self.set_palette_layout(
            {QtCore.Qt.LeftDockWidgetArea: QParamExploreInspector,
             QtCore.Qt.RightDockWidgetArea: QParameterView,
             })
            
    def set_action_links(self):
        self.action_links = \
            {
              'execute': ('explore_changed', self.explore_non_empty),
            }

    def set_action_defaults(self):
        self.action_defaults.update(
            { 'execute': [('setEnabled', True, self.set_execute_action),
                          ('setIcon', False, CurrentTheme.EXECUTE_EXPLORE_ICON),
                          ('setToolTip', False, 'Execute the parameter exploration')],
              'publishWeb': [('setEnabled', False, False)],
              'publishPaper': [('setEnabled', False, False)],
            })
            
    def set_execute_action(self):
        if self.controller and self.controller.vistrail:
            versionId = self.controller.current_version
            return self.controller.vistrail.get_paramexp(versionId) is not None
        return False
    
    def explore_non_empty(self, on):
        return on
    
    def exploreChange(self, on):
        from vistrails.gui.vistrails_window import _app
        _app.notify('explore_changed', on)
        
    def execute(self):
        """ execute() -> None        
        Perform the exploration by collecting a list of actions
        corresponding to each dimension
        
        """
        # persist the parameter exploration
        pe = self.getParameterExploration()
        pe.action_id = self.controller.current_version

        # check if pe has changed
        changed = False
        if not self.controller.current_parameter_exploration or \
         pe != self.controller.current_parameter_exploration:
            changed = True
            pe.name = ''
            self.controller.current_parameter_exploration = pe
            self.controller.vistrail.add_paramexp(pe)
            self.controller.set_changed(True)
        else:
            pe = self.controller.current_parameter_exploration

        errors = self.controller.executeParameterExploration(pe,
                                     self.get_param_view().pipeline_view.scene())
        if errors:
            errors = '\n'.join(['Position %s: %s' % (error[0], error[1]) for error in errors])
            debug.critical("Parameter Exploration Execution had errors", errors)
        if changed:
            from vistrails.gui.vistrails_window import _app
            _app.notify('exploration_changed')



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from vistrails.core.modules.utils import make_modules_dict

try:
    # write_numpy requires numpy
    import numpy
except ImportError: # pragma: no cover
    numpy_modules = []
else:
    from .write_numpy import _modules as numpy_modules

from .write_excel import get_xlwt
if get_xlwt():
    from .write_excel import _modules as excel_modules
else: # pragma: no cover
    excel_modules = []

from .write_csv import _modules as csv_modules


_modules = make_modules_dict(numpy_modules, csv_modules, excel_modules,
                             namespace='write')


###############################################################################

import unittest


class BaseWriteTestCase(object):
    def test_numeric(self):
        from vistrails.tests.utils import execute, intercept_result
        from ..common import ExtractColumn
        from ..identifiers import identifier
        with intercept_result(ExtractColumn, 'value') as results:
            self.assertFalse(execute([
                    ('BuildTable', identifier, [
                        ('a', [('List', '[1, 2, 3]')]),
                        ('b', [('List', '[4, 5, 6]')]),
                    ]),
                    (self.WRITER_MODULE, identifier, []),
                    (self.READER_MODULE, identifier, []),
                    ('ExtractColumn', identifier, [
                        ('column_index', [('Integer', '1')]),
                        ('numeric', [('Boolean', 'True')]),
                    ]),
                ], [
                    (0, 'value', 1, 'table'),
                    (1, 'file', 2, 'file'),
                    (2, 'value', 3, 'table'),
                ],
                add_port_specs=[
                    (0, 'input', 'a',
                     'org.vistrails.vistrails.basic:List'),
                    (0, 'input', 'b',
                     'org.vistrails.vistrails.basic:List'),
                ]))
        self.assertEqual(len(results), 1)
        self.assertEqual(list(results[0]), [4, 5, 6])

    def test_strings(self):
        from vistrails.tests.utils import execute, intercept_result
        from ..common import ExtractColumn
        from ..identifiers import identifier
        with intercept_result(ExtractColumn, 'value') as results:
            self.assertFalse(execute([
                    ('BuildTable', identifier, [
                        ('a', [('List', "['a', '2', 'c']")]),
                        ('b', [('List', "[4, 5, 6]")]),
                    ]),
                    (self.WRITER_MODULE, identifier, []),
                    (self.READER_MODULE, identifier, []),
                    ('ExtractColumn', identifier, [
                        ('column_index', [('Integer', '0')]),
                        ('numeric', [('Boolean', 'False')]),
                    ]),
                ], [
                    (0, 'value', 1, 'table'),
                    (1, 'file', 2, 'file'),
                    (2, 'value', 3, 'table'),
                ],
                add_port_specs=[
                    (0, 'input', 'a',
                     '(org.vistrails.vistrails.basic:List)'),
                    (0, 'input', 'b',
                     '(org.vistrails.vistrails.basic:List)'),
                ]))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0], ['a', '2', 'c'])


@unittest.skipIf(get_xlwt() is None, "xlwt not available")
class ExcelWriteTestCase(unittest.TestCase, BaseWriteTestCase):
    WRITER_MODULE = 'write|WriteExcelSpreadsheet'
    READER_MODULE = 'read|ExcelSpreadsheet'


class CSVWriteTestCase(unittest.TestCase, BaseWriteTestCase):
    WRITER_MODULE = 'write|WriteCSV'
    READER_MODULE = 'read|CSVFile'



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
from vistrails.db.versions.v0_9_4.domain import DBVistrail, DBAction, DBTag, DBModule, \
    DBConnection, DBPortSpec, DBFunction, DBParameter, DBLocation, DBAdd, \
    DBChange, DBDelete, DBAnnotation, DBPort, DBGroup, \
    DBWorkflow, DBLog, DBAbstraction

def update_workflow(old_obj, translate_dict):
    return DBWorkflow.update_version(old_obj.db_workflow, 
                                     translate_dict, DBWorkflow())

def update_modules(old_obj, trans_dict):
    new_modules = []
    for obj in old_obj.db_modules:
        if obj.vtType == 'module':
            new_modules.append(DBModule.update_version(obj, trans_dict))
        elif obj.vtType == 'abstractionRef':
            new_modules.append(DBAbstraction.update_version(obj,
                                                               trans_dict))
        elif obj.vtType == 'group':
            new_modules.append(DBGroup.update_version(obj, trans_dict))
    return new_modules

def translateVistrail(_vistrail):
    def update_operations(old_obj, trans_dict):
        def update_abstractionRef(old_obj, trans_dict):
            def get_internal_version(old_obj, trans_dict):
                return str(old_obj.db_version)
            def get_version(old_obj, trans_dict):
                return ''
            new_dict = {'DBAbstraction':
                        {'internal_version': get_internal_version,
                         'version': get_version}}
            new_dict.update(trans_dict)
            return DBAbstraction.update_version(old_obj.db_data, new_dict)
        new_ops = []
        for obj in old_obj.db_operations:
            if obj.vtType == 'add':
                if obj.db_what == 'abstractionRef':
                    trans_dict['DBAdd'] = {'data': update_abstractionRef}
                    new_op = DBAdd.update_version(obj, trans_dict)
                    new_op.db_what = 'abstraction'
                    new_ops.append(new_op)
                    del trans_dict['DBAdd']
                else:
                    new_op = DBAdd.update_version(obj, trans_dict)
                    if obj.db_parentObjType == 'abstractionRef':
                        new_op.db_parentObjType = 'abstraction'
                    new_ops.append(new_op)
            elif obj.vtType == 'delete':
                new_ops.append(DBDelete.update_version(obj, trans_dict))
            elif obj.vtType == 'change':
                if obj.db_what == 'abstractionRef':
                    trans_dict['DBChange'] = {'data': update_abstractionRef}
                    new_op = DBChange.update_version(obj, trans_dict)
                    new_op.db_what = 'abstraction'
                    new_ops.append(new_op)
                    del trans_dict['DBChange']
                else:
                    new_op = DBChange.update_version(obj, trans_dict)
                    if obj.db_parentObjType == 'abstractionRef':
                        new_op.db_parentObjType = 'abstraction'
                    new_ops.append(new_op)
        return new_ops

    translate_dict = {'DBGroup': {'workflow': update_workflow},
                      'DBAction': {'operations': update_operations},
                      'DBWorkflow': {'modules': update_modules}}
    vistrail = DBVistrail.update_version(_vistrail, translate_dict)
    vistrail.db_version = '0.9.4'
    return vistrail

def translateWorkflow(_workflow):
    translate_dict = {'DBGroup': {'workflow': update_workflow},
                      'DBWorkflow': {'modules': update_modules}}
    workflow = DBWorkflow.update_version(_workflow, translate_dict)
    workflow.db_version = '0.9.4'
    return workflow

def translateLog(_log):
    translate_dict = {}
    log = DBLog.update_version(_log, translate_dict)
    log.db_version = '0.9.4'
    return log
